import os
import sys
import json
import logging
import traceback
from pprint import pprint
from requests.exceptions import HTTPError, ConnectionError
from deriva.core import __version__ as VERSION, read_config, init_logging, format_exception, BaseCLI
from deriva.core.utils import eprint

CLIENT_CRED_FILE = '/home/secrets/oauth2/client_secret_globus.json'


class UsageException(ValueError):
    """Usage exception.
    """

    def __init__(self, message):
        """Initializes the exception.
        """
        super(UsageException, self).__init__(message)


class GlobusAuthAPIException(Exception):
    pass


GlobusAuthAPIError = GlobusAuthAPIException


class GlobusAuthUtil:
    def __init__(self, **kwargs):
        client_id = kwargs.get("client_id")
        client_secret = kwargs.get("client_secret")
        if not (client_id and client_secret):
            cred_file = kwargs.get("credential_file", CLIENT_CRED_FILE)
            creds = read_config(cred_file)
            client_id = creds['web'].get('client_id')
            client_secret = creds['web'].get('client_secret')
        try:
            from globus_sdk import ConfidentialAppAuthClient
            global GlobusAuthAPIError
            from globus_sdk.exc import AuthAPIError as GlobusAuthAPIError
            self.client = ConfidentialAppAuthClient(client_id, client_secret)
            self.client_id = client_id
        except Exception as e:
            logging.error("Unable to instantiate required globus_sdk class ConfidentialAppAuthClient: %s" % e)
            raise

    @staticmethod
    def from_json(obj):
        # if "obj" is a dict, use it. Otherwise, see if it's a JSON string or a file containing json.
        if isinstance(obj, dict):
            return obj
        try:
            result = json.loads(obj)
        except ValueError:
            # if this fails, we have nothing left to try, so don't bother catching errors
            f = open(obj, 'r')
            result = json.load(f)
            f.close()
        return result

    def list_all_scopes(self):
        r = self.client.get("/v2/api/scopes")
        return r.data

    def list_scope(self, scope):
        r = self.client.get("/v2/api/scopes/{s}".format(s=scope))
        return r.data

    def create_scope(self, scope):
        if not scope:
            raise UsageException("A supported scope argument is required.")

        r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id),
                             json_body=self.from_json(scope))
        return r.data

    def update_scope(self, scope_id, scope):
        if not (scope_id and scope):
            raise UsageException("The scope_id and scope arguments are required.")

        r = self.client.put("/v2/api/scopes/{scope_id}".format(scope_id=scope_id),
                            json_body=self.from_json(scope))
        return r.data

    def add_fqdn_to_client(self, fqdn):
        if not fqdn:
            raise UsageException("A fqdn (fully qualified domain name) argument is required.")
        r = self.client.post('/v2/api/clients/{client_id}/fqdns'.format(client_id=self.client_id),
                             json_body={'fqdn': fqdn})
        return r.data

    def get_client_for_fqdn(self, fqdn):
        if not fqdn:
            raise UsageException("A fqdn (fully qualified domain name) argument is required.")
        r = self.client.get('/v2/api/clients?fqdn={fqdn}'.format(fqdn=fqdn))
        return r.data

    def get_client(self, client_id):
        r = self.client.get('/v2/api/clients/{client_id}'.format(client_id=client_id if client_id else self.client_id))
        return r.data

    def get_clients(self):
        r = self.client.get('/v2/api/clients')
        return r.data

    def verify_access_token(self, token):
        if not token:
            raise UsageException("A token argument is required.")
        r = self.client.oauth2_validate_token(token)
        return r.data

    def introspect_access_token(self, token):
        if not token:
            raise UsageException("A token argument is required.")
        r = self.client.oauth2_token_introspect(token)
        return r.data

    def new_client(self, client):
        r = self.client.post("/v2/api/clients",
                             json_body=self.from_json(client))
        return r.data

    def create_client(self,
                      name,
                      redirect_uris,
                      public=False,
                      visibility="private",
                      project=None,
                      required_idp=None,
                      preselect_idp=None,
                      terms_of_service=None,
                      privacy_policy=None):
        if not (name and redirect_uris):
            raise UsageException("The name and redirect_uris arguments are required.")
        client = {
            "client": {
                "name": name,
                "public_client": public,
                "redirect_uris": redirect_uris,
                "visibility": visibility if (visibility == "private" or visibility == "public") else "private"
            }
        }
        if project:
            client["client"].update({"project": project})
        if required_idp:
            client["client"].update({"required_idp": required_idp})
        if preselect_idp:
            client["client"].update({"preselect_idp": preselect_idp})
        if terms_of_service or privacy_policy:
            client["client"].update({"links": {"terms_of_service": terms_of_service,
                                               "privacy_policy": privacy_policy}})
        return self.new_client(client)

    def update_client(self, client, client_id=None):
        if not client:
            raise UsageException("A client argument is required.")

        r = self.client.put("/v2/api/clients/{client_id}".format(client_id=client_id if client_id else self.client_id),
                            json_body=self.from_json(client))
        return r.data

    def delete_client(self, client_id):
        if not client_id:
            raise UsageException("A client argument is required.")

        r = self.client.delete("/v2/api/clients/{client_id}".format(client_id=client_id))
        return r.data

    def add_redirect_uris(self, redirect_uris):
        if not redirect_uris:
            raise UsageException("A redirect_uris argument is required.")
        d = {
            "client": {
                "redirect_uris": redirect_uris
            }
        }
        return self.update_client(d)

    def get_my_client(self):
        r = self.client.get('/v2/api/clients/{client_id}'.format(client_id=self.client_id))
        return r.data

    def get_scopes_by_name(self, scope_name):
        if not scope_name:
            raise UsageException("A scope_name argument is required.")
        scopes = self.client.get('/v2/api/scopes?scope_strings={sname}'.format(sname=scope_name))
        if scopes is None:
            return None
        else:
            return scopes.get("scopes")

    def get_scopes_by_id(self, id_string):
        if not id_string:
            raise UsageException("A id_string argument is required.")
        scopes = self.client.get('/v2/api/scopes?ids={ids}'.format(ids=id_string))
        if scopes is None:
            return None
        else:
            return scopes.get("scopes")

    def my_scope_ids(self):
        c = self.client.get('/v2/api/clients/{client_id}'.format(client_id=self.client_id))
        me = c.get("client")
        if me is None or me.get('scopes') is None:
            return []
        else:
            return me.get('scopes')

    def my_scope_names(self):
        scope_names = []
        scope_ids = self.my_scope_ids()
        if scope_ids is not None:
            ids = ",".join(scope_ids)
            print(str(ids))
            scopes = self.get_scopes_by_id(ids)
            for scope in scopes:
                scope_names.append(scope.get('scope_string'))
        return scope_names

    def get_grant_types(self):
        grant_types = None
        c = self.client.get('/v2/api/clients/{client_id}'.format(client_id=self.client_id))
        me = c.get("client")
        if me is not None:
            grant_types = me.get('grant_types')
        return grant_types

    def add_scopes(self, new_scopes):
        if not new_scopes:
            raise UsageException("A new_scopes argument is required.")
        scopes = set(self.my_scope_ids())
        for scope in self.get_scopes_by_name(",".join(new_scopes)):
            scopes.add(scope.get('id'))
        d = {
            "client": {
                "scopes": list(scopes)
            }
        }

        r = self.client.put('/v2/api/clients/{client_id}'.format(client_id=self.client_id), json_body=d)
        return r.data

    def add_dependent_scopes(self, parent_scope_name, child_scope_names, optional=False, requires_refresh_token=False):
        if not (parent_scope_name and child_scope_names):
            raise UsageException("The parent_scope_name and child_scope_names arguments are required.")

        child_scope_ids = set()
        parent_scopes = self.get_scopes_by_name(parent_scope_name)
        if parent_scopes is None:
            return "no parent scope"
        if len(parent_scopes) != 1:
            return "{sl} parent scopes: {p}".format(sl=str(len(parent_scopes)), p=str(parent_scopes))
        parent_scope_id = parent_scopes[0].get("id")
        for scope in parent_scopes[0].get('dependent_scopes'):
            child_scope_ids.add(scope.get('id'))
        new_child_scopes = self.get_scopes_by_name(",".join(child_scope_names))
        for scope in new_child_scopes:
            child_scope_ids.add(scope.get('id'))
        dependent_scopes = []
        for scope_id in child_scope_ids:
            dependent_scopes.append({'scope': scope_id,
                                     'optional': optional,
                                     'requires_refresh_token': requires_refresh_token})
        d = {
            "scope": {
                "dependent_scopes": dependent_scopes
            }
        }
        print(str(d))
        r = self.client.put('/v2/api/scopes/{i}'.format(i=parent_scope_id),
                            json_body=d)
        return r.data

    def create_scope_with_deps(self, name, description, suffix, dependent_scopes=[], advertised=True,
                               allow_refresh_tokens=True):
        if not (name and description and suffix):
            raise UsageException("The name, description and suffix arguments are required.")
        dependent_scope_arg = []
        if len(dependent_scopes) > 0:
            child_scopes = self.get_scopes_by_name(",".join(dependent_scopes))
            for scope in child_scopes:
                dependent_scope_arg.append({
                    "scope": scope.get("id"),
                    "optional": False,
                    "requires_refresh_token": False
                })
        scope = {
            "scope": {
                "name": name,
                "description": description,
                "scope_suffix": suffix,
                "dependent_scopes": dependent_scope_arg,
                "advertised": advertised,
                "allows_refresh_tokens": allow_refresh_tokens
            }
        }

        r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id),
                             json_body=scope)
        return r.data

    def delete_scope(self, scope_name):
        if not scope_name:
            raise UsageException("A scope_name argument is required.")
        scopes = self.get_scopes_by_name(scope_name)
        if scopes is None or len(scopes) != 1:
            return "null or multiple scopes"
        scope_id = scopes[0].get('id')
        if scope_id is None:
            return "no scope id"
        r = self.client.delete('/v2/api/scopes/{scope_id}'.format(scope_id=scope_id))
        return r.data

    def get_dependent_scopes(self, scope):
        if not scope:
            raise UsageException("A supported scope argument is required.")
        result = {"scope_string": scope.get("scope_string"), "dependent_scopes": []}
        for ds in scope.get("dependent_scopes"):
            ds_id = ds.get('scope')
            ds_info = {"id": ds_id}
            d = self.get_scopes_by_id(ds_id)
            if d is not None:
                ds_info['scope_string'] = d[0].get('scope_string')
            result['dependent_scopes'].append(ds_info)
        return result


class DerivaGlobusAuthUtilCLIException(Exception):
    def __init__(self, message):
        super(DerivaGlobusAuthUtilCLIException, self).__init__(message)


class DerivaGlobusAuthUtilCLI(BaseCLI):
    """Deriva GlobusClientUtil Command-line Interface.
    """

    def __init__(self, *args, **kwargs):
        super(DerivaGlobusAuthUtilCLI, self).__init__(*args, **kwargs)

        self.gau = None
        self.remove_options(['--host', '--config-file', '--credential-file', '--token', '--oauth2-token'])
        self.parser.add_argument("--pretty", "-p", action="store_true",
                                 help="Pretty-print all result output.")
        parent_mutex_group = self.parser.add_mutually_exclusive_group(required=True)
        parent_mutex_group.add_argument('--credential-file', '-f', metavar='<file>',
                                        help="Path to a credential file.")
        parent_mutex_group.add_argument('--client-id', '-c', metavar='<client id>',
                                        help="Globus Auth Client ID")
        self.parser.add_argument('--client-secret', '-k', metavar='<client secret key>',
                                 help="Globus Auth Client Secret")

        subparsers = self.parser.add_subparsers(title='sub-commands', dest='subcmd')

        # get-scope parser
        get_scopes = subparsers.add_parser('get-scopes',
                                           help="Get one or more scopes by name or ID, or all scopes owned by this "
                                                "client if no scope ID is specified.")
        get_scopes.add_argument("--scope-ids", metavar="[scopes]",
                                help="A comma-delimited list of scope IDs to retrieve.")
        get_scopes.add_argument("--scope-names", metavar="[scopes]",
                                help="A comma-delimited list of scope names to retrieve.")
        get_scopes.set_defaults(func=self.get_scopes)

        # put-scope parser
        put_scope = subparsers.add_parser("put-scope",
                                          help="Create or update a scope owned by this client from a valid JSON scope "
                                               "configuration in string form, or a path to an equivalent file.")
        put_scope.add_argument("--scope-id", metavar="<scope id>", help="The scope ID.")
        put_scope.add_argument("scope_config", metavar="<JSON String or path to file>",
                               help="A valid JSON scope configuration in string form, or a path to an equivalent file.")
        put_scope.set_defaults(func=self.put_scope)

        # add-scope parser
        add_scopes = subparsers.add_parser('add-scopes',
                                           help="Add one or more scopes by name to this client, or add one or more "
                                                "scopes as dependent scopes of a specified parent scope.")
        add_scopes.add_argument("scope_names", metavar="[scopes]", default=list(),
                                type=lambda s: [item for item in s.split(',')],
                                help="A comma-delimited list of scope names.")
        add_scopes.add_argument("--parent-scope", metavar="<scope name>",
                                help="The parent scope name if adding dependent scopes.")
        add_scopes.add_argument("--optional", action="store_true",
                                help="When adding to dependent scopes, sets the scope as optional. Default false.")
        add_scopes.add_argument("--requires-refresh-token", action="store_true",
                                help="When adding to dependent scopes, sets the scope to require the usage of refresh "
                                     "tokens. Default false.")
        add_scopes.set_defaults(func=self.add_scopes)

        # create-scope parser
        create_scope = subparsers.add_parser('create-scope',
                                             help="Creates a scope for each registered FQDN + the id of the client "
                                                  "from command-line parameters.")
        create_scope.add_argument("name", metavar="<scope name>",
                                  help="The new scope name. Max 100 chars.")
        create_scope.add_argument("description", metavar="<scope desc>",
                                  help="A scope description. Max 5000 chars.")
        create_scope.add_argument("suffix", metavar="<scope suffix>",
                                  help="String consisting of lowercase letters, number, and underscores.")
        create_scope.add_argument("--dependent-scope-names", metavar="[scopes]", default=list(),
                                  type=lambda s: [item for item in s.split(',')],
                                  help="A comma-delimited list of dependent scope names.")
        create_scope.add_argument("--advertised", action="store_true",
                                  help="Whether or not the scope should show up in searches. Default: True")
        create_scope.add_argument("--allow-refresh-token", action="store_true",
                                  help="Whether or not the scope allows refresh tokens to be issued. Default: True")
        create_scope.set_defaults(func=self.create_scope)

        # del-scope parser
        del_scope = subparsers.add_parser("del-scope",
                                          help="Delete the specified scope name. "
                                               "Deleting a scope deletes all resources associated with it. This "
                                               "operation can cause other apps and services that depend on the scope "
                                               "to stop working. This action cannot be undone.")
        del_scope.add_argument("scope_name", metavar="<scope_name>", help="The scope name to delete.")
        del_scope.set_defaults(func=self.delete_scope)

        # get-client parser
        get_client = subparsers.add_parser('get-client',
                                           help="Retrieve client information for the specified client, "
                                                "or all clients owned by this client ID if no ID specified.")
        get_client.add_argument("--get-client-id", metavar="<client ID>",
                                help="Retrieve client information for the specified client ID.")
        get_client.set_defaults(func=self.get_client)

        # put-client parser
        put_client = subparsers.add_parser('put-client',
                                           help="Update this (or another) client OR create a child client of this "
                                                "client from a valid JSON client configuration in string form, or a "
                                                "path to an equivalent file.")
        put_client.add_argument("client_config", metavar="<JSON String or path to file>",
                                help="A valid JSON client config in string form, or a path to an equivalent file.")
        put_client.add_argument("--put-client-id", metavar="<client id>",
                                help="The client ID to update, or implicitly this client's ID if not specified.")
        put_client.add_argument("--create", action="store_true",
                                help="Create a new child client from the input client config.")
        put_client.set_defaults(func=self.put_client)

        # create-client parser
        create_client = subparsers.add_parser('create-client',
                                              help="Create a client from command-line parameters.")
        create_client.add_argument("name", metavar="<client name>",
                                   help="Display name shown to users in consents. String without line-breaks, with no "
                                        "more than 100 characters.")
        create_client.add_argument("redirect_uris", metavar="[redirect uris]", default=list(),
                                   type=lambda s: [item for item in s.split(',')],
                                   help="A comma-delimited list of URIs that may be used in OAuth authorization flows.")
        create_client.add_argument("--public", action="store_true",
                                   help="Create a public (native app) client.")
        create_client.add_argument("--visibility", choices=["public", "private"],
                                   help="\"private\" means that only entities in the same project as the client "
                                        "can view it. \"public\" means that any authenticated entity can view it.")
        create_client.add_argument("--project", metavar="<project ID>",
                                   help="ID representing the project this client belongs to.")
        create_client.add_argument("--required-idp", metavar="<IDP ID>",
                                   help="ID representing an Identity Provider. In order to use this client a user must"
                                        " have an identity from this IdP in their identity set.")
        create_client.add_argument("--preselect-idp", metavar="<IDP ID>",
                                   help="ID representing an Identity Provider. This preselects the given IdP on the "
                                        "Globus Auth login page if the user is not already authenticated.")
        create_client.add_argument("--terms-of-service", metavar="<URL>",
                                   help="A URL to the terms and conditions statement for this client.")
        create_client.add_argument("--privacy_policy", metavar="<URL>",
                                   help="A URL to the privacy policy for this client.")
        create_client.set_defaults(func=self.create_client)

        # delete-client parser
        delete_client = subparsers.add_parser('del-client',
                                              help="Delete a client by ID. Warning: deletes all resources associated "
                                                   "with it the client. This includes user consents, scopes (which "
                                                   "means this operation can cause other apps and services that depend "
                                                   "on those scopes to stop working as well), and any child clients "
                                                   "owned by the client (which in turn means that all resources "
                                                   "associated with the child clients would get deleted as well). "
                                                   "This action cannot be undone.")
        delete_client.add_argument("del_client_id", metavar="<client id>",
                                   help="The client ID to update, or implicitly this client's ID if not specified.")
        delete_client.set_defaults(func=self.delete_client)

        # client-fqdn parser
        client_fqdn = subparsers.add_parser('client-fqdn',
                                            help="Retrieve client information for an FQDN, "
                                                 "or add and FQDN to this client.")
        client_fqdn.add_argument("--add", action="store_true",
                                 help="Add the specified FQDN to this client ID")
        client_fqdn.add_argument("fqdn", metavar="<fqdn>",
                                 help="The fully qualified domain name to lookup or add to this client.")
        client_fqdn.set_defaults(func=self.client_fqdn)

        # token parser
        token = subparsers.add_parser('token',
                                      help="Introspect or validate an access token.")
        token.add_argument("--validate", action="store_true",
                           help="Validate the access token.")
        token.add_argument("token", metavar="<token>",
                           help="The access token to introspect (or validate).")
        token.set_defaults(func=self.token)

    def get_scopes(self, args):
        if args.scope_ids:
            return self.gau.get_scopes_by_id(args.scope_ids)
        elif args.scope_names:
            return self.gau.get_scopes_by_name(args.scope_names)
        else:
            return self.gau.list_all_scopes()

    def put_scope(self, args):
        if args.scope_id:
            return self.gau.update_scope(args.scope_config)
        else:
            return self.gau.create_scope(args.scope_config)

    def add_scopes(self, args):
        if args.parent_scope:
            return self.gau.add_dependent_scopes(args.parent_scope,
                                                 args.scope_names,
                                                 args.optional,
                                                 args.requires_refresh_token)
        else:
            return self.gau.add_scopes(args.scope_names)

    def create_scope(self, args):
        return self.gau.create_scope_with_deps(args.name,
                                               args.description,
                                               args.suffix,
                                               args.dependent_scope_names,
                                               args.advertised,
                                               args.allow_refresh_token)

    def delete_scope(self, args):
        return self.gau.delete_scope(args.scope_name)

    def get_client(self, args):
        if args.get_client_id:
            return self.gau.get_client(args.get_client_id)
        else:
            return self.gau.get_clients()

    def put_client(self, args):
        if args.create:
            return self.gau.new_client(args.client_config)
        else:
            return self.gau.update_client(args.client_config, args.put_client_id)

    def create_client(self, args):
        return self.gau.create_client(args.name,
                                      args.redirect_uris,
                                      args.public,
                                      args.visibility,
                                      args.project,
                                      args.required_idp,
                                      args.preselect_idp,
                                      args.terms_of_service,
                                      args.privacy_policy)

    def delete_client(self, args):
        return self.gau.delete_client(args.del_client_id)

    def client_fqdn(self, args):
        if args.add:
            return self.gau.add_fqdn_to_client(args.fqdn)
        else:
            return self.gau.get_client_for_fqdn(args.fqdn)

    def token(self, args):
        if args.validate:
            return self.gau.verify_access_token(args.token)
        else:
            return self.gau.introspect_access_token(args.token)

    def main(self):
        args = self.parse_cli()

        def _cmd_error_message(emsg):
            return "{prog} {subcmd}: {msg}".format(
                prog=self.parser.prog, subcmd=args.subcmd, msg=emsg)

        try:
            if not hasattr(args, 'func'):
                self.parser.print_usage()
                return 1

            self.gau = GlobusAuthUtil(**vars(args))
            response = args.func(args)
            if args.pretty:
                if isinstance(response, dict) or isinstance(response, list):
                    try:
                        print(json.dumps(response, indent=2))
                    except:
                        pprint(response)
                else:
                    pprint(response)
            else:
                print(response)
            return 0

        except UsageException as e:
            eprint("{prog} {subcmd}: {msg}".format(prog=self.parser.prog, subcmd=args.subcmd, msg=e))
        except ConnectionError as e:
            eprint("{prog}: Connection error occurred".format(prog=self.parser.prog))
        except GlobusAuthAPIError as e:
            if 401 == e.http_status:
                msg = 'Authentication required: %s' % e.message
            elif 403 == e.http_status:
                msg = 'Permission denied: %s' % e.message
            else:
                msg = e
            logging.debug(format_exception(e))
            eprint(_cmd_error_message(msg))
        except RuntimeError as e:
            logging.debug(format_exception(e))
            eprint('Unexpected runtime error occurred')
        except:
            eprint('Unexpected error occurred')
            traceback.print_exc()
        return 1


def main():
    desc = "DERIVA Globus Auth Utilities"
    info = "For more information see: https://github.com/informatics-isi-edu/deriva-py"
    return DerivaGlobusAuthUtilCLI(desc, info).main()


if __name__ == '__main__':
    sys.exit(main())

# if __name__ == '__main__':
#    scope_file = sys.argv[1]
#    token = sys.argv[1]
# s = GlobusClientUtil()
# s.add_dependent_scopes('https://auth.globus.org/scopes/0fb084ec-401d-41f4-990e-e236f325010a/deriva_test_withdeps',
#                        ['openid',
#                         'email',
#                         'urn:globus:auth:scope:nexus.api.globus.org:groups',
#                         'https://auth.globus.org/scopes/identifiers.globus.org/create_update'
#                         ])
# print s.add_grant_types([
#     "openid",
#     "email",
#     "profile",
#     "urn:globus:auth:scope:auth.globus.org:view_identities",
#     "urn:globus:auth:scope:nexus.api.globus.org:groups",
#     "https://auth.globus.org/scopes/identifiers.globus.org/create_update"
# ])
#    s.add_scopes(["openid", "email"])
#    print str(s.my_scope_names())
#    print s.update_private_client()
#    pprint.pprint(s.get_scopes_by_name('email,urn:globus:auth:scope:nexus.api.globus.org:groups,urn:globus:auth:scope:transfer.api.globus.org:all'))
#    print s.create_private_client("nih_test_3", ["https://webauthn-dev.isrd.isi.edu/authn/session", "https://nih-commons.derivacloud.org/authn/session"])
#    print s.get_clients()
#    print s.add_scopes(]
#    print s.get_my_client()
#    print s.add_redirect_uris(["https://webauthn-dev.isrd.isi.edu/authn/session", "https://nih-commons.derivacloud.org/authn/session"])
#    print s.create_scope(scope_file)
#    print s.add_fqdn_to_client('nih-commons.derivacloud.org')
# print s.create_scope_with_deps('Deriva Services', 'Use Deriva Services', 'deriva_all',
#                                dependent_scopes = [
#                                    "openid",
#                                    "email",
#                                    "profile",
#                                    "urn:globus:auth:scope:auth.globus.org:view_identities",
#                                    "urn:globus:auth:scope:nexus.api.globus.org:groups",
#                                    "urn:globus:auth:scope:transfer.api.globus.org:all",
#                                    "https://auth.globus.org/scopes/identifiers.globus.org/create_update"
#                                    ])
# print s.delete_scope("https://auth.globus.org/scopes/nih-commons.derivacloud.org/deriva_test_nodeps")
# print s.delete_scope("https://auth.globus.org/scopes/0fb084ec-401d-41f4-990e-e236f325010a/deriva_test_withdeps")
# print s.delete_scope("https://auth.globus.org/scopes/nih-commons.derivacloud.org/deriva_test_withdeps")
# print s.delete_scope("https://auth.globus.org/scopes/0fb084ec-401d-41f4-990e-e236f325010a/deriva_test_3")
# print s.delete_scope("https://auth.globus.org/scopes/nih-commons.derivacloud.org/deriva_test_3")
# print s.delete_scope("https://auth.globus.org/scopes/0fb084ec-401d-41f4-990e-e236f325010a/deriva_test_4")
# print s.delete_scope("https://auth.globus.org/scopes/nih-commons.derivacloud.org/deriva_test_4")
# print(str(s.list_all_scopes()))
#    scope = s.get_scopes_by_name("https://auth.globus.org/scopes/0fb084ec-401d-41f4-990e-e236f325010a/deriva_all")[0]
#    pprint.pprint(s.get_dependent_scopes(scope))
#    print s.verify_access_token(token)
#    print s.introspect_access_token(token)
# print s.update_scope('23b9a3f9-872d-4a40-9c4c-a80a4c61f3bf',
#                      {"name" : "Use Deriva Services",
#                       "description" : "Use all Deriva services"
#                       })
# print s.update_scope('b892c8a9-2f33-4404-9fe3-6eb9093010c3',
#                      {"name" : "Use Deriva Services on nih-commons.derivacloud.org",
#                       "description" : "Use all Deriva services on nih-commons.derivacloud.org"
#                       })



from . import urlquote
from datetime import date
import logging
import re
from requests import HTTPError
import warnings

logger = logging.getLogger(__name__)
"""Logger for this module"""

_system_defaults = {'RID', 'RCT', 'RCB', 'RMT', 'RMB'}
"""Set of system default column names"""


def _kwargs(**kwargs):
    """Helper for extending datapath with sub-types for the whole model tree."""
    kwargs2 = {
        'schema_class': Schema,
        'table_class': Table,
        'column_class': Column
    }
    kwargs2.update(kwargs)
    return kwargs2


def from_catalog(catalog):
    """Creates a datapath.Catalog object from an ErmrestCatalog object.
    :param catalog: an ErmrestCatalog object
    :return: a datapath.Catalog object
    """
    return Catalog(catalog.getCatalogSchema(), **_kwargs(catalog=catalog))


def _isidentifier(a):
    """Tests if string is a valid python identifier.
    This function is intended for internal usage within this module.
    :param a: a string
    """
    if hasattr(a, 'isidentifier'):
        return a.isidentifier()
    else:
        return re.match("[_A-Za-z][_a-zA-Z0-9]*$", a) is not None


def _http_error_message(e):
    """Returns a formatted error message from the raw HTTPError.
    """
    return '\n'.join(e.response.text.splitlines()[1:]) + '\n' + str(e)


class DataPathException (Exception):
    """DataPath exception
    """
    def __init__(self, message, reason=None):
        super(DataPathException, self).__init__(message, reason)
        self.message = message
        self.reason = reason

    def __str__(self):
        return self.message


class Catalog (object):
    """Handle to a Catalog.
    """
    def __init__(self, model_doc, **kwargs):
        """Creates the Catalog.
        :param model_doc: the schema document for the catalog
        """
        super(Catalog, self).__init__()
        self.schemas = {
            sname: kwargs.get('schema_class', Schema)(sname, sdoc, **kwargs)
            for sname, sdoc in model_doc.get('schemas', {}).items()
        }

    def __dir__(self):
        return dir(Catalog) + ['schemas'] + [key for key in self.schemas if _isidentifier(key)]

    def __getattr__(self, a):
        if a in self.schemas:
            return self.schemas[a]
        else:
            return getattr(super(Catalog, self), a)


class Schema (object):
    """Represents a Schema.
    """
    def __init__(self, sname, schema_doc, **kwargs):
        """Creates the Schema.
        :param sname: the schema's name
        :param schema_doc: the schema document
        """
        super(Schema, self).__init__()
        self.name = sname
        self.tables = {
            tname: kwargs.get('table_class', Table)(sname, tname, tdoc, **kwargs)
            for tname, tdoc in schema_doc.get('tables', {}).items()
        }

    def __dir__(self):
        return dir(Schema) + ['name', 'tables'] + [key for key in self.tables if _isidentifier(key)]

    def __getattr__(self, a):
        if a in self.tables:
            return self.tables[a]
        else:
            return getattr(super(Schema, self), a)

    def describe(self):
        """Provides a description of the model element.

        :return: a user-friendly string representation of the model element.
        """
        s = "Schema name: '%s'\nList of tables:\n" % self.name
        if len(self.tables) == 0:
            s += "none"
        else:
            s += "\n".join("  '%s'" % tname for tname in self.tables)
        return s

    def _repr_html_(self):
        return self.describe()


class DataPath (object):
    """Represents an arbitrary data path."""
    def __init__(self, root):
        assert isinstance(root, TableAlias)
        self._path_expression = Root(root)
        self._root = root
        self._base_uri = root.catalog._server_uri
        self._table_instances = dict()  # map of alias_name => TableAlias object
        self._context = None
        self._identifiers = []
        self._bind_table_instance(root)

    def __dir__(self):
        return dir(DataPath) + self._identifiers

    def __getattr__(self, a):
        if a in self._table_instances:
            return self._table_instances[a]
        else:
            return getattr(super(DataPath, self), a)

    @property
    def table_instances(self):
        return self._table_instances

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        assert isinstance(value, TableAlias)
        assert value.name in self._table_instances
        if self._context != value:
            self._path_expression = ResetContext(self._path_expression, value)
            self._context = value

    @property
    def uri(self):
        return self._base_uri + str(self._path_expression)

    def _contextualized_uri(self, context):
        """Returns a path uri for the specified context.

        :param context: a table instance that is bound to this path
        :return: string representation of the path uri
        """
        assert isinstance(context, TableAlias)
        assert context.name in self._table_instances
        if self._context != context:
            return self._base_uri + str(ResetContext(self._path_expression, context))
        else:
            return self.uri

    def _bind_table_instance(self, alias):
        """Binds a new table instance into this path.
        """
        assert isinstance(alias, TableAlias)
        alias.path = self
        self._table_instances[alias.name] = self._context = alias
        if _isidentifier(alias.name):
            self._identifiers.append(alias.name)

    def delete(self):
        """Deletes the entity set referenced by the data path.
        """
        try:
            path = str(self._path_expression)
            logger.debug("Deleting: {p}".format(p=path))
            self._root.catalog.delete(path)
        except HTTPError as e:
            logger.debug(e.response.text)
            if 400 <= e.response.status_code < 500:
                raise DataPathException(_http_error_message(e), e)
            else:
                raise e

    def filter(self, filter_expression):
        """Filters the path based on the specified formula.

        :param filter_expression: should be a valid Predicate object
        :return: self
        """
        assert isinstance(filter_expression, Predicate)
        self._path_expression = Filter(self._path_expression, filter_expression)
        return self

    def link(self, right, on=None, join_type=''):
        """Links this path with another table.

        To link a table with an unambigious relationship where table A is related to table B via a single foreign key
        reference, the `on` clause is not.

        ```
        # let A and B be variables for tables from the catalog
        path = A.link(B)
        ```

        To link tables with more than one foreign key reference between them, use explicit `on` clause.

        ```
        # let A.c1 be a column that is a simple foreign key to B.c1 that is a simple key in B
        path = A.link(B, on=(A.c1 == B.c1))
        ```

        To link tables with foreign keys on composite keys, use a conjunction of 2 or more equality comparisons in the
        `on` clause.

        ```
        # let A.c1, A.c2 be columns that form a foreign key to B.c1, B.c2 that are a composite key in B
        path = A.link(B, on=((A.c1 == B.c1) & (A.c2 == B.c2)))
        ```

        By default links use inner join semantics on the foreign key / key equality comparison. The `join_type`
        parameter can be used to specify `left`, `right`, or `full` outer join semantics.

        :param right: the right hand table of the link expression
        :param on: an equality comparison between key and foreign key columns or a conjunction of such comparisons
        :param join_type: the join type of this link which may be 'left', 'right', 'full' outer joins or '' for inner
        join link by default.
        :return: self
        """
        if not isinstance(right, Table):
            raise ValueError("'right' must be a 'Table' instance")
        if on and not (isinstance(on, ComparisonPredicate) or (isinstance(on, ConjunctionPredicate) and
                                                               on.is_valid_join_condition)):
            raise ValueError("'on' must be a comparison or conjuction of comparisons")
        if join_type and on is None:
            raise ValueError("'on' must be specified for outer joins")
        if right.catalog != self._root.catalog:
            raise ValueError("'right' is from a different catalog. Cannot link across catalogs.")
        if isinstance(right, TableAlias) and right.name in self._table_instances:
            raise ValueError("'right' is a table alias that has already been used.")
        else:
            # Generate an unused alias name for the table
            table_name = right.name
            alias_name = table_name
            counter = 1
            while alias_name in self._table_instances:
                counter += 1
                alias_name = table_name + str(counter)
            right = right.alias(alias_name)

        if on is None:
            on = right

        # Extend path expression
        self._path_expression = Link(self._path_expression, on, right, join_type)

        # Bind alias and this data path
        self._bind_table_instance(right)

        return self

    def entities(self, *attributes, **renamed_attributes):
        """Returns a results set of whole entities from this data path's current context.

        ```
        results1 = my_path.entities()
        results2 = my_path.entities(col1, col2)  # WARNING. Deprecated usage. Use the `attributes(...)` method instead.
        ```

        :param attributes: DEPRECATED.
        :param renamed_attributes: DEPRECATED.
        :return: a result set of entities where each element is a whole entity per the table definition and policy.
        """
        if attributes or renamed_attributes:
            warnings.warn("Use of 'attributes' or 'renamed_attributes' in 'entities(...)' is deprecated. Use 'attributes(...)' instead.", DeprecationWarning)
        return self._query(attributes, renamed_attributes)

    def aggregates(self, **functions):
        """Returns a results set of computed aggregates from this data path.

        By using the built-in subclasses of the `AggregateFunction` class, including `Min`, `Max`, `Avg`, `Cnt`, `CntD`,
        `Array`, and `ArrayD`, aggregates can be computed and fetched. These aggregates must be passed as named
        parameters since they require _alias names_.

        ```
        results1 = my_path.aggregates(min_col1=Min(col1), arr_col2=Array(col2))
        results2 = my_path.aggregates(Min(col1), Array(col2))  # Error! Aggregates must be named.
        results3 = my_path.aggregates(col1, arr_col2=Array(col2))  # Error! Cannot mix columns and aggregate functions.
        ```

        :param functions: named parameters of type AggregateFunction
        :return: a results set with a single row of results.
        """
        return self._query([], functions)

    def attributegroups(self, group_key, **functions):
        """Returns a results set of computed aggregates for groups of attributes from this data path.

        Aggregates over groups, as specified by a `group_key`, can be computed and fetched. Note that the `group_key`
        named parameter is therefore _reserved_ for any invocation of the `attributegroups(...)` method.

        ```
        results1 = my_path.attributegroups(group_key=col1, min_col1=Min(col2), arr_col2=Array(col3))  # 1 group key
        results2 = my_path.attributegroups(group_key=(col1, col2), min_col1=Min(col3), arr_col2=Array(col4))  # >1 group keys
        ```

        As with aggregation, callers must not mix ordinary columns in with grouped aggregates.

        :param group_key: a Column or a set of Columns to be used as the group key for grouping the computations on.
        :param functions: named parameters of type AggregateFunction
        :return: a results set with a row of results for each group.
        """
        functions['group_key'] = group_key
        return self._query([], functions)

    def attributes(self, *attributes, **renamed_attributes):
        """Returns a results set of attributes projected and optionally renamed from this data path.

        ```
        results1 = my_path.attributes(col1, col2)  # fetch a subset of attributes of the path
        results2 = my_path.attributes(foo=col1, bar=col2)  # fetch and rename the attributes
        results3 = my_path.attributes(col1, bar=col2)  # rename some but not others
        ```

        :param attributes: a list of Columns.
        :param renamed_attributes: a list of renamed Columns.
        :return: a results set of the projected attributes from this data path.
        """
        return self._query(attributes, renamed_attributes)

    def _query(self, attributes, renamed_attributes, context=None):
        """Internal method for querying the data path from the perspective of the given 'context'.

        :param attributes: a list of Columns.
        :param renamed_attributes: a list of renamed Columns or AggregateFunctions
        :param context: optional context for the query.
        :return: a results set.
        """
        assert context is None or isinstance(context, TableAlias)
        catalog = self._root.catalog

        expression = self._path_expression
        if context:
            expression = ResetContext(expression, context)
        if attributes or renamed_attributes:
            expression = Project(expression, attributes, renamed_attributes)
        base_path = str(expression)

        def fetcher(limit=None, sort=None):
            assert limit is None or isinstance(limit, int)
            assert sort is None or hasattr(sort, '__iter__')
            limiting = '?limit=%d' % limit if limit else ''
            sorting = '@sort(' + ','.join([col.uname for col in sort]) + ')' if sort else ''
            path = base_path + sorting + limiting
            logger.debug("Fetching " + path)
            try:
                resp = catalog.get(path)
                return resp.json()
            except HTTPError as e:
                logger.debug(e.response.text)
                if 400 <= e.response.status_code < 500:
                    raise DataPathException(_http_error_message(e), e)
                else:
                    raise e

        return ResultSet(self._base_uri + base_path, fetcher)


class ResultSet (object):
    """A set of results for various queries or data manipulations.

    The ResultSet is produced by a path. The results may be explicitly fetched. The ResultSet behaves like a
    container. If the ResultSet has not been fetched explicitly, on first use of container operations, it will
    be implicitly fetched from the catalog.
    """
    def __init__(self, uri, fetcher_fn):
        """Initializes the ResultSet.
        :param uri: the uri for the entity set in the catalog.
        :param fetcher_fn: a function that fetches the entities from the catalog.
        """
        assert fetcher_fn is not None
        self._fetcher_fn = fetcher_fn
        self._results_doc = None
        self._dataframe = None
        self.uri = uri

    @property
    def _results(self):
        if self._results_doc is None:
            self.fetch()
        return self._results_doc

    @property
    def dataframe(self):
        """Pandas DataFrame representation of this path.

        :return: a pandas dataframe object
        :raise ImportError: exception if the 'pandas' library is not available
        """
        if not self._dataframe:
            from pandas import DataFrame
            self._dataframe = DataFrame(self._results)
        return self._dataframe

    def __len__(self):
        return len(self._results)

    def __getitem__(self, item):
        return self._results[item]

    def __iter__(self):
        return iter(self._results)

    def fetch(self, limit=None, sort=None):
        """Fetches the results from the catalog.

        :param limit: maximum number of results to fetch from the catalog.
        :param sort: collection of columns to use for sorting. Note that if using the `attributes(...)` query interface,
        the columns in the sort list _must_ be explicitly listed in the `attributes(...)` call.
        :return: self
        """
        limit = int(limit) if limit else None
        self._results_doc = self._fetcher_fn(limit, sort)
        self._dataframe = None  # clear potentially cached state
        logger.debug("Fetched %d entities" % len(self._results_doc))
        return self


class Table (object):
    """Represents a Table.
    """
    def __init__(self, sname, tname, table_doc, **kwargs):
        """Creates a Table object.
        :param sname: name of the schema
        :param tname: name of the table
        :param table_doc: deserialized json document of the table definition
        :param kwargs: must include `catalog`
        """
        self.catalog = kwargs['catalog']
        self.sname = sname
        self.name = tname
        self._table_doc = table_doc
        self._kwargs = kwargs

        kwargs.update(table=self)
        self.column_definitions = {
            cdoc['name']: kwargs.get('column_class', Column)(sname, tname, cdoc, **kwargs)
            for cdoc in table_doc.get('column_definitions', [])
        }

    def __dir__(self):
        return dir(Table) + ['catalog', 'sname', 'name'] + [key for key in self.column_definitions if _isidentifier(key)]

    def __getattr__(self, a):
        if a in self.column_definitions:
            return self.column_definitions[a]
        else:
            return getattr(super(Table, self), a)

    def describe(self):
        """Provides a description of the model element.

        :return: a user-friendly string representation of the model element.
        """
        s = "Table name: '%s'\nList of columns:\n" % self.name
        if len(self.column_definitions) == 0:
            s += "none"
        else:
            s += "\n".join("  %s" % col.name for col in self.column_definitions.values())
        return s

    def _repr_html_(self):
        return self.describe()

    @property
    def uname(self):
        """the url encoded name"""
        return urlquote(self.name)

    @property
    def fqname(self):
        """the url encoded fully qualified name"""
        return "%s:%s" % (urlquote(self.sname), self.uname)

    @property
    def instancename(self):
        return '*'

    @property
    def fromname(self):
        return self.fqname

    @property
    def path(self):
        """Always a new DataPath instance that is rooted at this table.
        Note that this table will be automatically aliased using its own table name.
        """
        return DataPath(self.alias(self.name))

    @path.setter
    def path(self, value):
        """Not allowed on base tables.
        """
        raise Exception("Path assignment not allowed on base table objects.")

    @property
    def _contextualized_path(self):
        """Returns the path as contextualized for this table instance.

        Conditionally updates the context of the path to which this table instance is bound.
        """
        return self.path

    @property
    def uri(self):
        return self.path.uri

    def alias(self, alias_name):
        """Returns a table alias object.
        :param alias_name: a string to use as the alias name
        """
        return TableAlias(self, alias_name)

    def filter(self, filter_expression):
        """See the docs for this method in `DataPath` for more information."""
        return self._contextualized_path.filter(filter_expression)

    def link(self, right, on=None, join_type=''):
        """See the docs for this method in `DataPath` for more information."""
        return self._contextualized_path.link(right, on, join_type)

    def _query(self, attributes, renamed_attributes):
        """Invokes query on the path for this table."""
        return self.path._query(attributes, renamed_attributes)

    def entities(self, *attributes, **renamed_attributes):
        """Returns a results set of whole entities from this data path's current context.

        See the docs for this method in `DataPath` for more information.
        """
        if attributes or renamed_attributes:
            warnings.warn("Use of 'attributes' or 'renamed_attributes' in 'entities(...)' is deprecated. Use 'attributes(...)' instead.", DeprecationWarning)
        return self._query(attributes, renamed_attributes)

    def aggregates(self, **functions):
        """Returns a results set of computed aggregates from this data path.

        See the docs for this method in `DataPath` for more information.
        """
        return self._query([], functions)

    def attributegroups(self, group_key, **functions):
        """Returns a results set of computed aggregates for groups of attributes from this data path.

        See the docs for this method in `DataPath` for more information.
        """
        functions['group_key'] = group_key
        return self._query([], functions)

    def attributes(self, *attributes, **renamed_attributes):
        """Returns a results set of attributes projected and optionally renamed from this data path.

        See the docs for this method in `DataPath` for more information.
        """
        return self._query(attributes, renamed_attributes)

    def insert(self, entities, defaults=set(), nondefaults=set(), add_system_defaults=True):
        """Inserts entities into the table.
        :param entities: an iterable collection of entities (i.e., rows) to be inserted into the table.
        :param defaults: optional, set of column names to be assigned the default expression value.
        :param nondefaults: optional, set of columns names to override implicit system defaults
        :param add_system_defaults: flag to add system columns to the set of default columns.
        :return a ResultSet of newly created entities.
        """
        # empty entities will be accepted but results are therefore an empty entity set
        if not entities:
            return ResultSet(self.path.uri, lambda ignore1, ignore2: [])

        options = []

        if defaults or add_system_defaults:
            defaults_enc = {urlquote(cname) for cname in defaults}
            if add_system_defaults:
                defaults_enc |= _system_defaults - nondefaults
            options.append("defaults={cols}".format(cols=','.join(defaults_enc)))

        if nondefaults:
            nondefaults_enc = {urlquote(cname) for cname in nondefaults}
            options.append("nondefaults={cols}".format(cols=','.join(nondefaults_enc)))

        path = '/entity/' + self.fqname
        if options:
            path += "?" + "&".join(options)
        logger.debug("Inserting entities to path: {path}".format(path=path))

        # JSONEncoder does not handle general iterable objects, so we have to make sure its an acceptable collection
        if not hasattr(entities, '__iter__'):
            raise ValueError('entities is not iterable')
        entities = entities if isinstance(entities, (list, tuple)) else list(entities)

        # test the first entity element to make sure that it looks like a dictionary
        if not hasattr(entities[0], 'keys'):
            raise ValueError('entities[0] does not look like a dictionary -- does not have a "keys()" method')

        try:
            resp = self.catalog.post(path, json=entities, headers={'Content-Type': 'application/json'})
            return ResultSet(self.path.uri, lambda ignore1, ignore2: resp.json())
        except HTTPError as e:
            logger.debug(e.response.text)
            if 400 <= e.response.status_code < 500:
                raise DataPathException(_http_error_message(e), e)
            else:
                raise e

    def update(self, entities, correlation={'RID'}, targets=None):
        """Update entities of a table.

        For more information see the ERMrest protocol for the `attributegroup` interface. By default, this method will
        correlate the input data (entities) based on the `RID` column of the table. By default, the method will use all
        column names found in the first row of the `entities` input, which are not found in the `correlation` set and
        not defined as 'system columns' by ERMrest, as the targets if `targets` is not set.

        :param entities: an iterable collection of entities (i.e., rows) to be updated in the table.
        :param correlation: an iterable collection of column names used to correlate input set to the set of rows to be
        updated in the catalog. E.g., `{'col name'}` or `{mytable.mycolumn}` will work if you pass a Column object.
        :param targets: an iterable collection of column names used as the targets of the update operation.
        :return: a ResultSet of updated entities as returned by the corresponding ERMrest interface.
        """
        # empty entities will be accepted but results are therefore an empty entity set
        if not entities:
            return ResultSet(self.path.uri, lambda ignore1, ignore2: [])

        # JSONEncoder does not handle general iterable objects, so we have to make sure its an acceptable collection
        if not hasattr(entities, '__iter__'):
            raise ValueError('entities is not iterable')
        entities = entities if isinstance(entities, (list, tuple)) else list(entities)

        # test the first entity element to make sure that it looks like a dictionary
        if not hasattr(entities[0], 'keys'):
            raise ValueError('entities[0] does not look like a dictionary -- does not have a "keys()" method')

        # Form the correlation keys and the targets
        correlation_cnames = {urlquote(str(c)) for c in correlation}
        if targets:
            target_cnames = {urlquote(str(t)) for t in targets}
        else:
            exclusions = correlation_cnames | _system_defaults
            target_cnames = {urlquote(str(t)) for t in entities[0].keys() if urlquote(str(t)) not in exclusions}

        # test if there are any targets after excluding for correlation keys and system columns
        if not target_cnames:
            raise ValueError('No "targets" for the update. There must be at least one column as a target of the update,'
                             ' and targets cannot overlap with "correlation" keys and system columns.')

        # Form the path
        path = '/attributegroup/{table}/{correlation};{targets}'.format(
            table=self.fqname,
            correlation=','.join(correlation_cnames),
            targets=','.join(target_cnames)
        )

        try:
            resp = self.catalog.put(path, json=entities, headers={'Content-Type': 'application/json'})
            return ResultSet(self.path.uri, lambda ignore1, ignore2: resp.json())
        except HTTPError as e:
            logger.debug(e.response.text)
            if 400 <= e.response.status_code < 500:
                raise DataPathException(_http_error_message(e), e)
            else:
                raise e


class TableAlias (Table):
    """Represents a table alias.
    """
    def __init__(self, base_table, alias_name):
        """Initializes the table alias.

        :param base_table: the base table to be given an alias name
        :param alias_name: the alias name
        """
        assert isinstance(base_table, Table)
        super(TableAlias, self).__init__(base_table.sname, base_table.name, base_table._table_doc, **base_table._kwargs)
        self._base_table = base_table
        self.name = alias_name
        self._parent = None

    @property
    def uname(self):
        """the url encoded name"""
        return urlquote(self.name)

    @property
    def fqname(self):
        """the url encoded fully qualified name"""
        return self._base_table.fqname

    @property
    def instancename(self):
        return self.uname + ":*"

    @property
    def fromname(self):
        return "%s:=%s" % (self.uname, self._base_table.fqname)

    @property
    def path(self):
        """Returns the parent path for this alias.
        """
        if not self._parent:
            self._parent = DataPath(self)
        return self._parent

    @path.setter
    def path(self, value):
        if self._parent:
            raise Exception("Cannot bind a table instance that has already been bound.")
        elif not isinstance(value, DataPath):
            raise Exception("value must be a DataPath instance.")
        self._parent = value

    @property
    def _contextualized_path(self):
        """Returns the path as contextualized for this table instance.

        Conditionally updates the context of the path to which this table instance is bound.
        """
        path = self.path
        if path.context != self:
            path.context = self
        return path

    @property
    def uri(self):
        return self.path._contextualized_uri(self)

    def _query(self, attributes, renamed_attributes):
        """Overridden method to set context of query to this table instance."""
        return self.path._query(attributes, renamed_attributes, self)


class Column (object):
    """Represents a column in a table.
    """
    def __init__(self, sname, tname, column_doc, **kwargs):
        """Creates a Column object.
        :param sname: schema name
        :param tname: table name
        :param column_doc: column definition document
        :param kwargs: kwargs must include `table` a Table instance
        """
        super(Column, self).__init__()
        assert 'table' in kwargs
        assert isinstance(kwargs['table'], Table)
        self._table = kwargs['table']
        self.sname = sname
        self.tname = tname
        self.name = column_doc['name']
        self.type = column_doc['type']
        self.comment = column_doc['comment']

    def describe(self):
        """Provides a description of the model element.

        :return: a user-friendly string representation of the model element.
        """
        return "Column name: '%s'\tType: %s\tComment: '%s'" % \
               (self.name, self.type['typename'], self.comment)

    def _repr_html_(self):
        return self.describe()

    @property
    def uname(self):
        """the url encoded name"""
        return urlquote(self.name)

    @property
    def fqname(self):
        """the url encoded fully qualified name"""
        return "%s:%s" % (self._table.fqname, self.uname)

    @property
    def instancename(self):
        if isinstance(self._table, TableAlias):
            return "%s:%s" % (self._table.uname, self.uname)
        else:
            return self.uname

    @property
    def desc(self):
        """A descending sort modifier based on this column."""
        return SortDescending(self)

    def __str__(self):
        return self.name

    def eq(self, other):
        """Returns an 'equality' comparison predicate.

        :param other: `None` or any other literal value.
        :return: a filter predicate object
        """
        if other is None:
            return ComparisonPredicate(self, "::null::", '')
        else:
            return ComparisonPredicate(self, "=", other)

    __eq__ = eq

    def lt(self, other):
        """Returns a 'less than' comparison predicate.

        :param other: a literal value.
        :return: a filter predicate object
        """
        return ComparisonPredicate(self, "::lt::", other)

    __lt__ = lt

    def le(self, other):
        """Returns a 'less than or equal' comparison predicate.

        :param other: a literal value.
        :return: a filter predicate object
        """
        return ComparisonPredicate(self, "::leq::", other)

    __le__ = le

    def gt(self, other):
        """Returns a 'greater than' comparison predicate.

        :param other: a literal value.
        :return: a filter predicate object
        """
        return ComparisonPredicate(self, "::gt::", other)

    __gt__ = gt

    def ge(self, other):
        """Returns a 'greater than or equal' comparison predicate.

        :param other: a literal value.
        :return: a filter predicate object
        """
        return ComparisonPredicate(self, "::geq::", other)

    __ge__ = ge

    def regexp(self, other):
        """Returns a 'regular expression' comparison predicate.

        :param other: a _string_ literal value.
        :return: a filter predicate object
        """
        if not isinstance(other, str):
            logger.warning("'regexp' method comparison only supports string literals.")
        return ComparisonPredicate(self, "::regexp::", other)

    def ciregexp(self, other):
        """Returns a 'case-insensitive regular expression' comparison predicate.

        :param other: a _string_ literal value.
        :return: a filter predicate object
        """
        if not isinstance(other, str):
            logger.warning("'ciregexp' method comparison only supports string literals.")
        return ComparisonPredicate(self, "::ciregexp::", other)

    def ts(self, other):
        """Returns a 'text search' comparison predicate.

        :param other: a _string_ literal value.
        :return: a filter predicate object
        """
        if not isinstance(other, str):
            logger.warning("'ts' method comparison only supports string literals.")
        return ComparisonPredicate(self, "::ts::", other)


class SortDescending (object):
    """Represents a descending sort condition.
    """
    def __init__(self, col):
        """Creates sort descending object.

        :param col: a column object
        """
        assert isinstance(col, Column)
        self.col = col

    @property
    def uname(self):
        """the url encoded name"""
        return urlquote(self.col.uname) + "::desc::"


class PathOperator (object):
    def __init__(self, r):
        assert isinstance(r, PathOperator) or isinstance(r, Table)
        if isinstance(r, Project):
            raise Exception("Cannot extend a path after an attribute projection")
        self._r = r

    @property
    def _path(self):
        assert isinstance(self._r, PathOperator)
        return self._r._path

    @property
    def _mode(self):
        assert isinstance(self._r, PathOperator)
        return self._r._mode

    def __str__(self):
        return "/%s/%s" % (self._mode, self._path)


class Root (PathOperator):
    def __init__(self, r):
        super(Root, self).__init__(r)
        assert isinstance(r, Table)
        self._table = r

    @property
    def _path(self):
        return self._table.fromname

    @property
    def _mode(self):
        return 'entity'


class ResetContext (PathOperator):
    def __init__(self, r, alias):
        if isinstance(r, ResetContext):
            r = r._r  # discard the previous context reset operator
        super(ResetContext, self).__init__(r)
        assert isinstance(alias, TableAlias)
        self._alias = alias

    @property
    def _path(self):
        assert isinstance(self._r, PathOperator)
        return "%s/$%s" % (self._r._path, self._alias.uname)


class Filter(PathOperator):
    def __init__(self, r, formula):
        super(Filter, self).__init__(r)
        assert isinstance(formula, Predicate)
        self._formula = formula

    @property
    def _path(self):
        assert isinstance(self._r, PathOperator)
        return "%s/%s" % (self._r._path, str(self._formula))


class Project (PathOperator):
    def __init__(self, r, attributes, renamed_attributes):
        super(Project, self).__init__(r)
        assert len(attributes) > 0 or len(renamed_attributes) > 0
        self._attrs = []
        self._group_key = []
        self._inferred_mode = 'attribute'

        # Validate attributes
        for elem in attributes:
            if not (isinstance(elem, Column) or isinstance(elem, Table)):
                message = "Invalid value in attributes: %s." % str(elem)
                if isinstance(elem, AggregateFunction):
                    message += " Aggregate functions not allowed in attributes list, use renamed attributes instead."
                elif isinstance(elem, str):
                    message += " Try using the explicit syntax of \"path_or_table.column_definitions['your_column_name']\"."
                raise ValueError(message)

        # Validate renamed attributes
        for key in renamed_attributes:
            elem = renamed_attributes[key]
            if not (isinstance(elem, Column) or isinstance(elem, AggregateFunction)):
                message = "Invalid value in renamed attributes: %s=%s." % (str(key), str(elem))
                if isinstance(elem, str):
                    message += " Try using the explicit syntax of \"%s=path_or_table.column_definitions['your_column_name']\"." % key
                raise ValueError(message)

        # Validate group_key, if it exists
        if 'group_key' in renamed_attributes:
            group_key = renamed_attributes['group_key']
            del renamed_attributes['group_key']
            if isinstance(group_key, Column):
                self._group_key = [group_key.instancename]
            elif not all(isinstance(col, Column) for col in group_key):
                raise ValueError("Group keys must be Column objects.")
            else:
                self._group_key = [col.instancename for col in group_key]
            self._inferred_mode = 'attributegroup'

        # Validate generalized projection, if applicable
        aggregates = [isinstance(elem, AggregateFunction) for elem in renamed_attributes.values()]
        if any(aggregates):
            if not all(aggregates):
                raise ValueError("Aggregate functions must be used exclusively or not at all.")
            if not self._group_key:
                self._inferred_mode = 'aggregate'

        # Build the projection list
        self._attrs = [
            attr.instancename for attr in attributes
        ] + [
            "%s:=%s" % (urlquote(out_alias), attr.instancename) for out_alias, attr in renamed_attributes.items()
        ]

    @property
    def _path(self):
        assert isinstance(self._r, PathOperator)
        grouping = ','.join(self._group_key)
        projection = ','.join(self._attrs)
        if grouping and projection:
            return "%s/%s;%s" % (self._r._path, grouping, projection)
        else:
            return "%s/%s" % (self._r._path, grouping or projection)

    @property
    def _mode(self):
        return self._inferred_mode


class Link (PathOperator):
    def __init__(self, r, on, as_=None, join_type=''):
        super(Link, self).__init__(r)
        assert isinstance(on, ComparisonPredicate) or isinstance(on, Table) or (
                isinstance(on, ConjunctionPredicate) and on.is_valid_join_condition)
        assert as_ is None or isinstance(as_, TableAlias)
        assert join_type == '' or (join_type in ('left', 'right', 'full') and isinstance(on, Predicate))
        self._on = on
        self._as = as_
        self._join_type = join_type

    @property
    def _path(self):
        assert isinstance(self._r, PathOperator)
        assign = '' if self._as is None else "%s:=" % self._as.uname
        if isinstance(self._on, Table):
            cond = self._on.fqname
        elif isinstance(self._on, ComparisonPredicate):
            cond = str(self._on)
        elif isinstance(self._on, ConjunctionPredicate):
            cond = self._on.as_join_condition
        else:
            raise DataPathException("Invalid join condition: " + str(self._on))
        return "%s/%s%s%s" % (self._r._path, assign, self._join_type, cond)


class Predicate (object):
    """Common base class for all predicate types."""

    def and_(self, other):
        """Returns a conjunction predicate.

        :param other: a predicate object.
        :return: a junction predicate object.
        """
        if not isinstance(other, Predicate):
            raise ValueError("Invalid comparison with object that is not a Predicate instance.")
        return ConjunctionPredicate([self, other])

    __and__ = and_

    def or_(self, other):
        """Returns a disjunction predicate.

        :param other: a predicate object.
        :return: a junction predicate object.
        """
        if not isinstance(other, Predicate):
            raise ValueError("Invalid comparison with object that is not a Predicate instance.")
        return DisjunctionPredicate([self, other])

    __or__ = or_

    def negate(self):
        """Returns a negation predicate.

        This predicate is wrapped in a negation predicate which is returned to the caller.

        :return: a negation predicate object.
        """
        return NegationPredicate(self)

    __invert__ = negate


class ComparisonPredicate (Predicate):
    def __init__(self, lop, op, rop):
        super(ComparisonPredicate, self).__init__()
        assert isinstance(lop, Column)
        assert isinstance(rop, Column) or isinstance(rop, int) or \
            isinstance(rop, float) or isinstance(rop, str) or \
            isinstance(rop, date)
        assert isinstance(op, str)
        self._lop = lop
        self._op = op
        self._rop = rop

    @property
    def is_equality(self):
        return self._op == '='

    @property
    def left(self):
        return self._lop

    @property
    def right(self):
        return self._rop

    def __str__(self):
        if isinstance(self._rop, Column):
            # The only valid circumstance for a Column rop is in a link 'on' predicate for simple key/fkey joins
            return "(%s)=(%s)" % (self._lop.instancename, self._rop.fqname)
        else:
            # All other comparisons are serialized per the usual form
            return "%s%s%s" % (self._lop.instancename, self._op, urlquote(str(self._rop)))


class JunctionPredicate (Predicate):
    def __init__(self, op, operands):
        super(JunctionPredicate, self).__init__()
        assert operands and hasattr(operands, '__iter__') and len(operands) > 1
        assert all(isinstance(operand, Predicate) for operand in operands)
        assert isinstance(op, str)
        self._operands = operands
        self._op = op

    def __str__(self):
        return self._op.join(["(%s)" % operand for operand in self._operands])


class ConjunctionPredicate (JunctionPredicate):
    def __init__(self, operands):
        super(ConjunctionPredicate, self).__init__('&', operands)

    def and_(self, other):
        return ConjunctionPredicate(self._operands + [other])

    @property
    def is_valid_join_condition(self):
        """Tests if this conjunction is a valid join condition."""
        return all(isinstance(o, ComparisonPredicate) and o.is_equality for o in self._operands)

    @property
    def as_join_condition(self):
        """Returns the conjunction in the 'join condition' serialized format."""
        lhs = []
        rhs = []

        for operand in self._operands:
            assert isinstance(operand, ComparisonPredicate) and operand.is_equality
            assert isinstance(operand.left, Column)
            assert isinstance(operand.right, Column)
            lhs.append(operand.left)
            rhs.append(operand.right)

        return "({left})=({right})".format(
            left=",".join(lop.instancename for lop in lhs),
            right=",".join(rop.fqname for rop in rhs)
        )


class DisjunctionPredicate (JunctionPredicate):
    def __init__(self, operands):
        super(DisjunctionPredicate, self).__init__(';', operands)

    def or_(self, other):
        return DisjunctionPredicate(self._operands + [other])


class NegationPredicate (Predicate):
    def __init__(self, child):
        super(NegationPredicate, self).__init__()
        assert isinstance(child, Predicate)
        self._child = child

    def __str__(self):
        return "!(%s)" % self._child


class AggregateFunction (object):
    """Base class of all aggregate functions."""
    def __init__(self, name, operand):
        """Initializes the aggregate function.

        :param name: name of the function per ERMrest specification.
        :param operand: single operand of the function; a Column, Table, or TableAlias object.
        """
        super(AggregateFunction, self).__init__()
        self.name = name
        self.operand = operand

    def __str__(self):
        return "%s(%s)" % (self.name, self.operand)

    @property
    def instancename(self):
        return "%s(%s)" % (self.name, self.operand.instancename)


class Min (AggregateFunction):
    """Aggregate function for minimum non-NULL value."""
    def __init__(self, operand):
        super(Min, self).__init__('min', operand)


class Max (AggregateFunction):
    """Aggregate function for maximum non-NULL value."""
    def __init__(self, operand):
        super(Max, self).__init__('max', operand)


class Avg (AggregateFunction):
    """Aggregate function for average non-NULL value."""
    def __init__(self, operand):
        super(Avg, self).__init__('avg', operand)


class Cnt (AggregateFunction):
    """Aggregate function for count of non-NULL values."""
    def __init__(self, operand):
        super(Cnt, self).__init__('cnt', operand)


class CntD (AggregateFunction):
    """Aggregate function for count of distinct non-NULL values."""
    def __init__(self, operand):
        super(CntD, self).__init__('cnt_d', operand)


class Array (AggregateFunction):
    """Aggregate function for an array containing all values (including NULL)."""
    def __init__(self, operand):
        super(Array, self).__init__('array', operand)


class ArrayD (AggregateFunction):
    """Aggregate function for an array containing distinct values (including NULL)."""
    def __init__(self, operand):
        super(ArrayD, self).__init__('array_d', operand)
